@@ -3,7 +3,7 @@
     vdirsyncer.cli
     ~~~~~~~~~~~~~~
 
-    :copyright: (c) 2014 Markus Unterwaditzer
+    :copyright: (c) 2014 Markus Unterwaditzer & contributors
     :license: MIT, see LICENSE for more details.
 '''
 
@@ -14,8 +14,8 @@ import sys
 import argvard
 
 from .storage import storage_names
-from .sync import sync, StorageEmpty
-from .utils import expand_path, parse_options, split_dict
+from .sync import sync, StorageEmpty, SyncConflict
+from .utils import expand_path, parse_options, split_dict, get_class_init_args
 
 import vdirsyncer.log as log
 
@@ -29,6 +29,7 @@ except ImportError:
 cli_logger = log.get(__name__)
 
 PROJECT_HOME = 'https://github.com/untitaker/vdirsyncer'
+DOCS_HOME = 'https://vdirsyncer.readthedocs.org/en/latest'
 
 
 class CliError(RuntimeError):
@@ -110,6 +111,15 @@ def save_status(path, status_name, status):
             f.write('\n')
 
 
+def storage_class_from_config(config):
+    config = dict(config)
+    storage_name = config.pop('type')
+    cls = storage_names.get(storage_name, None)
+    if cls is None:
+        raise KeyError('Unknown storage: {}'.format(storage_name))
+    return cls, config
+
+
 def storage_instance_from_config(config, description=None):
     '''
     :param config: A configuration dictionary to pass as kwargs to the class
@@ -117,19 +127,18 @@ def storage_instance_from_config(config, description=None):
     :param description: A name for the storage for debugging purposes
     '''
 
-    config = dict(config)
-    storage_name = config.pop('type')
-    cls = storage_names[storage_name]
+    cls, config = storage_class_from_config(config)
+
     try:
         return cls(**config)
     except Exception:
-        all, required = get_init_args(cls)
+        all, required = get_class_init_args(cls)
         given = set(config)
         missing = required - given
         invalid = given - all
 
         cli_logger.critical('error: Failed to initialize {}'
-                            .format(description or storage_name))
+                            .format(description or cls.storage_name))
 
         if not missing and not invalid:
             cli_logger.exception('')
@@ -137,30 +146,28 @@ def storage_instance_from_config(config, description=None):
         if missing:
             cli_logger.critical(
                 u'error: {} storage requires the parameters: {}'
-                .format(storage_name, u', '.join(missing)))
+                .format(cls.storage_name, u', '.join(missing)))
 
         if invalid:
             cli_logger.critical(
                 u'error: {} storage doesn\'t take the parameters: {}'
-                .format(storage_name, u', '.join(invalid)))
+                .format(cls.storage_name, u', '.join(invalid)))
 
         sys.exit(1)
 
 
-def get_init_args(cls):
-    from vdirsyncer.storage.base import Storage
-    import inspect
-
-    if cls is Storage:
-        return set(), set()
-
-    spec = inspect.getargspec(cls.__init__)
-    all = set(spec.args[1:])
-    required = set(spec.args[1:-len(spec.defaults)])
-    supercls = next(x for x in cls.__mro__[1:] if hasattr(x, '__init__'))
-    s_all, s_required = get_init_args(supercls)
-
-    return all | s_all, required | s_required
+def expand_collection(pair, collection, all_pairs, all_storages):
+    if collection in ('from a', 'from b'):
+        a_name, b_name, _, storage_defaults = all_pairs[pair]
+        config = dict(storage_defaults)
+        if collection == 'from a':
+            config.update(all_storages[a_name])
+        else:
+            config.update(all_storages[b_name])
+        cls, config = storage_class_from_config(config)
+        return (s.collection for s in cls.discover(**config))
+    else:
+        return [collection]
 
 
 def main():
@@ -236,7 +243,7 @@ def _main(env, file_cfg):
     @sync_command.main('[pairs...]')
     def sync_main(context, pairs=None):
         '''
-        Syncronize the given pairs. If no pairs are given, all will be
+        Synchronize the given pairs. If no pairs are given, all will be
         synchronized.
 
         Examples:
@@ -246,28 +253,35 @@ def _main(env, file_cfg):
         from the pair "bob".
         '''
         actions = []
+        handled_collections = set()
         force_delete = context.get('force_delete', set())
-        for pair_name, collection in parse_pairs_args(pairs, all_pairs):
-            a_name, b_name, pair_options, storage_defaults = \
-                all_pairs[pair_name]
+        for pair_name, _collection in parse_pairs_args(pairs, all_pairs):
+            for collection in expand_collection(pair_name, _collection,
+                                                all_pairs, all_storages):
+                if (pair_name, collection) in handled_collections:
+                    continue
+                handled_collections.add((pair_name, collection))
 
-            config_a = dict(storage_defaults)
-            config_a['collection'] = collection
-            config_a.update(all_storages[a_name])
+                a_name, b_name, pair_options, storage_defaults = \
+                    all_pairs[pair_name]
 
-            config_b = dict(storage_defaults)
-            config_b['collection'] = collection
-            config_b.update(all_storages[b_name])
+                config_a = dict(storage_defaults)
+                config_a['collection'] = collection
+                config_a.update(all_storages[a_name])
 
-            actions.append({
-                'config_a': config_a,
-                'config_b': config_b,
-                'pair_name': pair_name,
-                'collection': collection,
-                'pair_options': pair_options,
-                'general': general,
-                'force_delete': force_delete
-            })
+                config_b = dict(storage_defaults)
+                config_b['collection'] = collection
+                config_b.update(all_storages[b_name])
+
+                actions.append({
+                    'config_a': config_a,
+                    'config_b': config_b,
+                    'pair_name': pair_name,
+                    'collection': collection,
+                    'pair_options': pair_options,
+                    'general': general,
+                    'force_delete': force_delete
+                })
 
         processes = general.get('processes', 0) or len(actions)
         cli_logger.debug('Using {} processes.'.format(processes))
@@ -312,13 +326,30 @@ def sync_collection(config_a, config_b, pair_name, collection, pair_options,
             force_delete=status_name in force_delete
         )
     except StorageEmpty as e:
-        side = 'a' if e.empty_storage is a else 'b'
-        storage = e.empty_storage
-        raise CliError(
-            '{collection_description}: Storage "{side}" ({storage}) was '
-            'completely emptied. Use "--force-delete {status_name}" to '
-            'synchronize that emptyness to the other side, or delete the '
-            'status by yourself to restore the items from the non-empty '
-            'side.'.format(**locals())
+        cli_logger.critical(
+            '{collection}: Storage "{side}" ({storage}) was completely '
+            'emptied. Use "--force-delete {status_name}" to synchronize that '
+            'emptyness to the other side, or delete the status by yourself to '
+            'restore the items from the non-empty side.'.format(
+                collection=collection_description,
+                side='a' if e.empty_storage is a else 'b',
+                storage=e.empty_storage,
+                status_name=status_name
+            )
         )
+    except SyncConflict as e:
+        cli_logger.critical(
+            '{collection}: One item changed on both sides. Resolve this '
+            'conflict manually, or by setting the `conflict_resolution` '
+            'parameter in your config file.\n'
+            'See also {docs}/api.html#pair-section\n'
+            'Item ID: {e.ident}\n'
+            'Item href on side A: {e.href_a}\n'
+            'Item href on side B: {e.href_b}\n'
+            .format(collection=collection_description, e=e, docs=DOCS_HOME)
+        )
+    except Exception:
+        cli_logger.exception('Unhandled exception occured while syncing {}.'
+                             .format(collection_description))
+
     save_status(general['status_path'], status_name, status)
