@@ -8,17 +8,32 @@
     interface is described in `vdirsyncer.storage.base`, the `Storage` class
     should be a superclass of all storage classes.
 
-    :copyright: (c) 2014 Markus Unterwaditzer
+    :copyright: (c) 2014 Markus Unterwaditzer & contributors
     :license: MIT, see LICENSE for more details.
 '''
 
 from .dav import CarddavStorage, CaldavStorage
 from .filesystem import FilesystemStorage
 from .http import HttpStorage
+from .singlefile import SingleFileStorage
 
-storage_names = {
-    'caldav': CaldavStorage,
-    'carddav': CarddavStorage,
-    'filesystem': FilesystemStorage,
-    'http': HttpStorage
-}
+
+def _generate_storage_dict(*classes):
+    rv = {}
+    for cls in classes:
+        key = cls.storage_name
+        assert key
+        assert isinstance(key, str)
+        assert key not in rv
+        rv[key] = cls
+    return rv
+
+storage_names = _generate_storage_dict(
+    CaldavStorage,
+    CarddavStorage,
+    FilesystemStorage,
+    HttpStorage,
+    SingleFileStorage
+)
+
+del _generate_storage_dict
